package ratelimiter

import (
	"sync"
	"time"
)

type Window struct {
	Start    time.Time
	Requests int
}

type IPRateLimiter struct {
	mu          sync.Mutex
	ips         map[string][]*Window
	windowSize  time.Duration // 滑动窗口时间长度
	maxRequests int           // 窗口内最大请求数
}

func NewIPRateLimiter(windowSize time.Duration, maxRequests int) *IPRateLimiter {
	return &IPRateLimiter{
		ips:         make(map[string][]*Window),
		windowSize:  windowSize,
		maxRequests: maxRequests,
	}
}

func (rl *IPRateLimiter) Allow(ip string) bool {
	rl.mu.Lock()
	defer rl.mu.Unlock()
	now := time.Now()
	cutoff := now.Add(-rl.windowSize)
	if windows, exists := rl.ips[ip]; exists {
		var validWindows []*Window
		for _, w := range windows {
			if w.Start.After(cutoff) {
				validWindows = append(validWindows, w)
			}
		}
		rl.ips[ip] = validWindows
	}
	total := 0
	for _, w := range rl.ips[ip] {
		total += w.Requests
	}
	if total >= rl.maxRequests {
		return false
	}
	if len(rl.ips[ip]) == 0 || now.Sub(rl.ips[ip][len(rl.ips[ip])-1].Start) > time.Second {
		rl.ips[ip] = append(rl.ips[ip], &Window{
			Start:    now,
			Requests: 1,
		})
	} else {
		rl.ips[ip][len(rl.ips[ip])-1].Requests++
	}
	return true
}
