package ws

import (
	"context"
	"encoding/json"
	"net/http"
	"sync"
	"time"

	ctxUtil "gitlab.wanzhuangkj.com/tush/xpkg/gin/ctxUtils"
	"gitlab.wanzhuangkj.com/tush/xpkg/jwt"
	"gitlab.wanzhuangkj.com/tush/xpkg/logger"
	"gitlab.wanzhuangkj.com/tush/xpkg/utils/xsf"
	"gitlab.wanzhuangkj.com/tush/xpkg/xerrors/xerror"

	"github.com/gin-gonic/gin"
	"github.com/gorilla/websocket"
	"go.uber.org/zap"
)

type Cmd string

type RedisChannel string

const (
	heartbeatDuration = 30 * time.Second
)

type wsConn struct {
	manager  *WsConnManager
	id       xsf.ID
	conn     *websocket.Conn
	sendBuff chan []byte
	log      *zap.Logger
	stop     chan struct{}
}

type WsConnManager struct {
	pool     map[xsf.ID]*wsConn
	upgrader *websocket.Upgrader
	sync.RWMutex
	log   *zap.Logger
	hooks []Hook
}

var (
	WsConnMgr *WsConnManager
)

type Hook interface {
	AfterSend(ctx context.Context, receiverID xsf.ID, payload *WsMessage) error
	AfterAccept(ctx context.Context, userID xsf.ID) error
	AfterUnregister(ctx context.Context, userID xsf.ID) error
}

func NewWsConnManager(logger *zap.Logger, upgrader *websocket.Upgrader, hooks ...Hook) *WsConnManager {
	return &WsConnManager{
		pool: make(map[xsf.ID]*wsConn),
		log:  logger, upgrader: upgrader, hooks: hooks}
}

func (m *WsConnManager) register(ctx context.Context, conn *wsConn) {
	m.Lock()
	defer m.Unlock()
	if oldClient, exists := m.pool[conn.id]; exists {
		oldClient.close(ctx)
	}
	m.pool[conn.id] = conn
}

// 注销连接
func (conn *wsConn) close(ctx context.Context) {
	conn.manager.Lock()
	defer conn.manager.Unlock()
	delete(conn.manager.pool, conn.id)
	close(conn.sendBuff)
	conn.stop <- struct{}{}
	close(conn.stop)
	conn.conn.Close()
	for _, h := range conn.manager.hooks {
		go func() {
			h.AfterUnregister(ctx, conn.id)
		}()
	}
}

// 处理 WebSocket 连接（示例 URL: ws://host/ws?userID=123&auth=xxxxxx）
func (m *WsConnManager) Accept(c *gin.Context) {
	err := m.accept(c)
	if err != nil {
		m.log.Error("handleWsConn failed", logger.Err(err))
	}
}

func (m *WsConnManager) accept(c *gin.Context) error {
	w, r := c.Writer, c.Request
	authorization := r.URL.Query().Get("auth")
	if authorization == "" {
		http.Error(w, "auth required", http.StatusUnauthorized)
		return xerror.New("auth required")
	}
	if len(authorization) <= 8 {
		return xerror.New("auth required")
	}
	token := authorization[7:] // remove Bearer prefix
	claims, err := jwt.ParseToken(token)
	if err != nil {
		return xerror.Wrap(err, "parse jwt")
	}
	userIDStr := claims.UID
	userID, _ := xsf.ParseString(userIDStr)
	conn, err := m.upgrader.Upgrade(w, r, nil)
	if err != nil {
		return xerror.Wrap(err, "ws upgrade failed")
	}
	client := &wsConn{
		id:       userID,
		conn:     conn,
		sendBuff: make(chan []byte, 256),
		manager:  m,
		log:      m.log,
		stop:     make(chan struct{}),
	}

	ctx := ctxUtil.WrapCtx(c)

	m.register(ctx, client)

	go client.heartBeat()
	go client.writeLoop()

	for _, h := range m.hooks {
		go func() {
			h.AfterAccept(ctx, userID)
		}()
	}
	return nil
}

func (m *WsConnManager) getConn(userID xsf.ID) (*wsConn, bool) {
	m.RLock()
	defer m.RUnlock()
	client, exists := m.pool[userID]
	return client, exists
}

func (m *WsConnManager) Send(ctx context.Context, receiverID xsf.ID, payload *WsMessage) error {
	if payload == nil {
		return nil
	}
	conn, exists := m.getConn(receiverID)
	if exists {
		data, _ := json.Marshal(payload)
		err := conn.send(ctx, data)
		if err != nil {
			return err
		}
	} else {
		return xerror.Newf("user[id:%d] ws conn not found", receiverID)
	}
	for _, h := range m.hooks {
		go func() {
			h.AfterSend(ctx, receiverID, payload)
		}()
	}
	return nil
}

func (c *wsConn) send(_ context.Context, msg []byte) error {
	// todo 1.  deadline time  2.  retry
	c.sendBuff <- msg
	return nil
}

func (c *wsConn) heartBeat() {
	ticker := time.NewTicker(heartbeatDuration)
	defer ticker.Stop()
	for {
		select {
		case <-ticker.C:
			if err := c.conn.WriteMessage(websocket.PingMessage, nil); err != nil {
				defer c.close(context.Background())
				return
			}
		case <-c.stop:
			return
		}
	}
}

func (c *wsConn) writeLoop() {
	for {
		select {
		case msg := <-c.sendBuff:
			if err := c.conn.WriteMessage(websocket.BinaryMessage, msg); err != nil {
				c.log.Error("write message failed", logger.Err(err))
			}
		case <-c.stop:
			return
		}
	}
}

const (
	Cmd_UserMessage = "user_message"
)

type WsMessage struct {
	Seq  string `json:"seq"`
	Cmd  Cmd    `json:"cmd"`
	Data any    `json:"payload"`
}
